from collections.abc import Awaitable, Iterable
from typing import Any, Callable

from litestar.connection import ASGIConnection
from litestar.handlers import BaseRouteHandler

from pyidentity.security.authorization import (
    AuthorizationHandlerContext,
    AuthorizationPolicyProvider,
    Authorize,
)

__all__ = ("authorize",)

from pyidentity.security.authorization import ClaimsAuthorize


def authorize(
    roles: Iterable[str] | str | None = None, policy: str | None = None
) -> Callable[[ASGIConnection, BaseRouteHandler], Awaitable[None]]:
    """Indicates that the route or router to which this dependency is applied requires the specified authorization.

    :param roles: A list of roles that are allowed to access the resource.
    :param policy: Policy name that determines access to the resource.
    :return:
    :raise InvalidOperationException: If the specified policy name is not found.
    :raise AuthorizationError: If authorization failed.
    """

    async def call_authorize(connection: ASGIConnection, _: BaseRouteHandler) -> None:
        await Authorize(roles, policy)(AuthorizationHandlerContext(connection), AuthorizationPolicyProvider())

    return call_authorize


def claims_authorize(
    claim_type: str, *allowed_values: Any, authentication_scheme: str | None = None
) -> Callable[[ASGIConnection, BaseRouteHandler], Awaitable[None]]:
    async def call_authorize(connection: ASGIConnection, _: BaseRouteHandler) -> None:
        await ClaimsAuthorize(claim_type, *allowed_values, authentication_scheme=authentication_scheme)(
            AuthorizationHandlerContext(connection), AuthorizationPolicyProvider()
        )

    return call_authorize
